"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getLatestCycleRecords = exports.cleanShardCycleData = exports.recordArchiversReputation = exports.getNewestCycleFromArchivers = exports.getNewestCycleFromConsensors = exports.fetchCycleRecords = exports.setShutdownCycleRecord = exports.validateCycleData = exports.validateCycle = exports.computeCycleMarker = exports.changeNetworkMode = exports.setLastProcessedMetaDataCounter = exports.setCurrentCycleMarker = exports.setCurrentCycleCounter = exports.setCurrentCycleDuration = exports.getCurrentCycleMarker = exports.getCurrentCycleCounter = exports.processCycles = exports.shardValuesByCycle = exports.currentNetworkMode = exports.cycleRecordWithShutDownMode = exports.removedAndApopedNodes = exports.CycleChain = exports.lastProcessedMetaData = exports.currentCycleDuration = void 0;
const NodeList = require("../NodeList");
const Crypto = require("../Crypto");
const State = require("../State");
const Logger = require("../Logger");
const P2P_1 = require("../P2P");
const profiler_1 = require("../profiler/profiler");
const nestedCounters_1 = require("../profiler/nestedCounters");
const cycleDataCache = require("../cache/cycleRecordsCache");
const ServiceQueue = require("../ServiceQueue");
const Data_1 = require("./Data");
const Utils = require("../Utils");
const Config_1 = require("../Config");
const node_fetch_1 = require("node-fetch");
const GossipData_1 = require("./GossipData");
const Collector_1 = require("./Collector");
const AccountDataProvider_1 = require("./AccountDataProvider");
const LostArchivers_1 = require("../LostArchivers");
const ShardFunctions_1 = require("../ShardFunctions");
const API_1 = require("../API");
const StringifyReduce_1 = require("../profiler/StringifyReduce");
const cycleRecordsCache_1 = require("../cache/cycleRecordsCache");
const cycles_1 = require("../dbstore/cycles");
const GlobalAccount_1 = require("../GlobalAccount");
const sync_v2_1 = require("../sync-v2");
exports.currentCycleDuration = 0;
let currentCycleCounter = -1;
let currentCycleMarker = '0'.repeat(32);
exports.lastProcessedMetaData = -1;
exports.CycleChain = new Map();
exports.removedAndApopedNodes = [];
exports.cycleRecordWithShutDownMode = null;
exports.currentNetworkMode = 'forming';
exports.shardValuesByCycle = new Map();
async function processCycles(cycles) {
    if (profiler_1.profilerInstance)
        profiler_1.profilerInstance.profileSectionStart('process_cycle', false);
    try {
        if (nestedCounters_1.nestedCountersInstance)
            nestedCounters_1.nestedCountersInstance.countEvent('cycle', 'process', 1);
        for (const cycle of cycles) {
            // Logger.mainLogger.debug('Current cycle counter', currentCycleCounter)
            // Skip if already processed [TODO] make this check more secure
            if (cycle.counter <= currentCycleCounter)
                continue;
            Logger.mainLogger.debug(new Date(), 'New Cycle received', cycle.counter);
            // Update currentCycle state
            exports.currentCycleDuration = cycle.duration * 1000;
            currentCycleCounter = cycle.counter;
            // Update NodeList from cycle info
            updateNodeList(cycle);
            updateNetworkTxsList(cycle);
            updateShardValues(cycle);
            changeNetworkMode(cycle.mode);
            (0, GossipData_1.getAdjacentLeftAndRightArchivers)();
            (0, LostArchivers_1.handleLostArchivers)(cycle);
            await (0, cycleRecordsCache_1.addCyclesToCache)(cycles);
            await (0, Collector_1.storeCycleData)([cycle]);
            Logger.mainLogger.debug(`Processed cycle ${cycle.counter}`);
            if (State.isActive) {
                (0, GossipData_1.sendDataToAdjacentArchivers)(GossipData_1.DataType.CYCLE, [cycle]);
                // Check the archivers reputaion in every new cycle & record the status
                recordArchiversReputation();
            }
            await (0, GlobalAccount_1.updateGlobalNetworkAccount)(cycle.counter);
            if (exports.currentNetworkMode === 'shutdown') {
                Logger.mainLogger.debug(Date.now(), `❌ Shutdown Cycle Record received at Cycle #: ${cycle.counter}`);
                await Utils.sleep(exports.currentCycleDuration);
                NodeList.clearNodeLists();
                await (0, Data_1.clearDataSenders)();
                setShutdownCycleRecord(cycle);
                NodeList.toggleFirstNode();
            }
            // Clean receipts/originalTxs cache that are older than <maxCyclesShardDataToKeep> minutes
            const cleanupTimestamp = (cycle.start - Config_1.config.maxCyclesShardDataToKeep * 60) * 1000;
            (0, Collector_1.cleanOldOriginalTxsMap)(cleanupTimestamp);
            (0, Collector_1.cleanOldReceiptsMap)(cleanupTimestamp);
            (0, exports.cleanShardCycleData)(cycle.counter - Config_1.config.maxCyclesShardDataToKeep);
        }
    }
    finally {
        if (profiler_1.profilerInstance)
            profiler_1.profilerInstance.profileSectionEnd('process_cycle', false);
    }
}
exports.processCycles = processCycles;
function getCurrentCycleCounter() {
    return currentCycleCounter;
}
exports.getCurrentCycleCounter = getCurrentCycleCounter;
function getCurrentCycleMarker() {
    return currentCycleMarker;
}
exports.getCurrentCycleMarker = getCurrentCycleMarker;
function setCurrentCycleDuration(duration) {
    exports.currentCycleDuration = duration * 1000;
}
exports.setCurrentCycleDuration = setCurrentCycleDuration;
function setCurrentCycleCounter(value) {
    currentCycleCounter = value;
}
exports.setCurrentCycleCounter = setCurrentCycleCounter;
function setCurrentCycleMarker(value) {
    currentCycleMarker = value;
}
exports.setCurrentCycleMarker = setCurrentCycleMarker;
function setLastProcessedMetaDataCounter(value) {
    exports.lastProcessedMetaData = value;
}
exports.setLastProcessedMetaDataCounter = setLastProcessedMetaDataCounter;
function changeNetworkMode(newMode) {
    if (newMode === exports.currentNetworkMode)
        return;
    // If the network mode is changed from restore to processing, clear the serving validators interval
    if (exports.currentNetworkMode === 'restore' && newMode === 'processing')
        (0, AccountDataProvider_1.clearServingValidatorsInterval)();
    if ((exports.currentNetworkMode === 'restart' || exports.currentNetworkMode === 'recovery') && newMode === 'restore') {
        NodeList.changeNodeListInRestore();
        (0, AccountDataProvider_1.initServingValidatorsInterval)();
    }
    if (exports.cycleRecordWithShutDownMode && newMode !== 'shutdown') {
        exports.cycleRecordWithShutDownMode = null;
    }
    exports.currentNetworkMode = newMode;
}
exports.changeNetworkMode = changeNetworkMode;
function computeCycleMarker(fields) {
    const cycleMarker = Crypto.hashObj(fields);
    return cycleMarker;
}
exports.computeCycleMarker = computeCycleMarker;
// validation of cycle record against previous marker
function validateCycle(prev, next) {
    const previousRecordWithoutMarker = { ...prev };
    delete previousRecordWithoutMarker.marker;
    const prevMarker = computeCycleMarker(previousRecordWithoutMarker);
    return next.previous === prevMarker;
}
exports.validateCycle = validateCycle;
const validateCycleData = (cycleRecord) => {
    const err = Utils.validateTypes(cycleRecord, {
        activated: 'a',
        activatedPublicKeys: 'a',
        active: 'n',
        apoptosized: 'a',
        archiverListHash: 's',
        counter: 'n',
        desired: 'n',
        duration: 'n',
        expired: 'n',
        joined: 'a',
        joinedArchivers: 'a',
        joinedConsensors: 'a',
        leavingArchivers: 'a',
        lost: 'a',
        lostSyncing: 'a',
        marker: 's',
        maxSyncTime: 'n',
        mode: 's',
        networkConfigHash: 's',
        networkId: 's',
        nodeListHash: 's',
        previous: 's',
        refreshedArchivers: 'a',
        refreshedConsensors: 'a',
        refuted: 'a',
        removed: 'a',
        returned: 'a',
        standbyAdd: 'a',
        standbyNodeListHash: 's',
        standbyRemove: 'a',
        start: 'n',
        syncing: 'n',
        target: 'n',
        archiversAtShutdown: 'a?',
        lostArchivers: 'a',
        refutedArchivers: 'a',
        removedArchivers: 'a',
    });
    if (err) {
        Logger.mainLogger.error('Invalid Cycle Record', err);
        return false;
    }
    const cycleRecordWithoutMarker = { ...cycleRecord };
    delete cycleRecordWithoutMarker.marker;
    if (computeCycleMarker(cycleRecordWithoutMarker) !== cycleRecord.marker) {
        Logger.mainLogger.error('Invalid Cycle Record: cycle marker does not match with the computed marker');
        return false;
    }
    return true;
};
exports.validateCycleData = validateCycleData;
function setShutdownCycleRecord(cycleRecord) {
    exports.cycleRecordWithShutDownMode = cycleRecord;
}
exports.setShutdownCycleRecord = setShutdownCycleRecord;
function updateNodeList(cycle) {
    const { joinedConsensors, activatedPublicKeys, removed, appRemoved, apoptosized, joinedArchivers, leavingArchivers, refreshedConsensors, refreshedArchivers, standbyAdd, standbyRemove, lostAfterSelection, } = cycle;
    const consensorInfos = joinedConsensors.map((jc) => ({
        ip: jc.externalIp,
        port: jc.externalPort,
        publicKey: jc.publicKey,
        id: jc.id,
    }));
    const refreshedConsensorInfos = refreshedConsensors.map((jc) => ({
        ip: jc.externalIp,
        port: jc.externalPort,
        publicKey: jc.publicKey,
        id: jc.id,
    }));
    NodeList.addNodes(NodeList.NodeStatus.SYNCING, consensorInfos);
    NodeList.setStatus(NodeList.NodeStatus.ACTIVE, activatedPublicKeys);
    NodeList.refreshNodes(NodeList.NodeStatus.ACTIVE, refreshedConsensorInfos);
    if (standbyAdd.length > 0) {
        const standbyNodeList = standbyAdd.map((joinRequest) => ({
            publicKey: joinRequest.nodeInfo.publicKey,
            ip: joinRequest.nodeInfo.externalIp,
            port: joinRequest.nodeInfo.externalPort,
        }));
        NodeList.addStandbyNodes(standbyNodeList);
    }
    if (standbyRemove.length > 0) {
        NodeList.removeStandbyNodes(standbyRemove);
    }
    const removedConsensusNodes = [];
    const removedPks = [...removed, ...appRemoved].reduce((keys, id) => {
        const nodeInfo = NodeList.getNodeInfoById(id);
        if (nodeInfo) {
            removedConsensusNodes.push(nodeInfo);
            keys.push(nodeInfo.publicKey);
        }
        return keys;
    }, []);
    NodeList.removeNodes(removedPks);
    const apoptosizedConsensusNodes = [];
    const apoptosizedPks = apoptosized.reduce((keys, id) => {
        const nodeInfo = NodeList.getNodeInfoById(id);
        if (nodeInfo) {
            apoptosizedConsensusNodes.push(nodeInfo);
            keys.push(nodeInfo.publicKey);
        }
        return keys;
    }, []);
    NodeList.removeNodes(apoptosizedPks);
    const lostAfterSelectionConsensusNodes = [];
    const lostAfterSelectionPks = lostAfterSelection.reduce((keys, id) => {
        const nodeInfo = NodeList.getNodeInfoById(id);
        if (nodeInfo) {
            lostAfterSelectionConsensusNodes.push(nodeInfo);
            keys.push(nodeInfo.publicKey);
        }
        return keys;
    }, []);
    NodeList.removeNodes(lostAfterSelectionPks);
    for (const joinedArchiver of joinedArchivers) {
        State.addArchiver(joinedArchiver);
    }
    for (const refreshedArchiver of refreshedArchivers) {
        State.addArchiver(refreshedArchiver);
    }
    for (const leavingArchiver of leavingArchivers) {
        State.removeActiveArchiver(leavingArchiver.publicKey);
    }
    const nodesToUnsubscribed = [...apoptosizedPks, ...removedPks];
    if (nodesToUnsubscribed.length > 0) {
        for (const key of nodesToUnsubscribed) {
            if (Data_1.dataSenders.has(key))
                (0, Data_1.unsubscribeDataSender)(key);
        }
    }
    if (removedConsensusNodes.length > 0 || apoptosizedConsensusNodes.length > 0) {
        exports.removedAndApopedNodes.push({
            cycle: cycle.counter,
            removed: removedConsensusNodes,
            apoptosized: apoptosizedConsensusNodes,
            lostAfterSelection: lostAfterSelectionConsensusNodes,
        });
        while (exports.removedAndApopedNodes.length > 10) {
            exports.removedAndApopedNodes.shift();
        }
    }
    NodeList.realUpdatedTimes.set('/nodelist', Date.now());
    NodeList.realUpdatedTimes.set('/full-nodelist', Date.now());
    // To pick nodes only when the archiver is active
    if (State.isActive) {
        (0, Data_1.subscribeConsensorsByConsensusRadius)();
    }
}
async function updateNetworkTxsList(cycle) {
    const { txadd, txremove } = cycle;
    ServiceQueue.addTxs(txadd);
    ServiceQueue.removeTxs(txremove);
    const calculatedTxListHash = ServiceQueue.getNetworkTxsListHash();
    if (calculatedTxListHash !== cycle.txlisthash) {
        console.error('txList hash from cycle record does not match the calculated txList hash');
        const syncTxListResult = await (0, sync_v2_1.syncTxList)(NodeList.getActiveList());
        syncTxListResult.match((txList) => {
            Logger.mainLogger.debug("Successfully synced txList from validators"),
                ServiceQueue.setTxList(txList);
        }, (error) => {
            Logger.mainLogger.error("Failed to synchronize transaction list:", error.message);
        });
    }
}
async function fetchCycleRecords(start, end) {
    const response = (await (0, API_1.queryFromArchivers)(API_1.RequestDataType.CYCLE, { start, end }));
    if (response)
        return response.cycleInfo;
    return [];
}
exports.fetchCycleRecords = fetchCycleRecords;
async function getNewestCycleFromConsensors(activeNodes) {
    const queryFn = async (node) => {
        const response = (await (0, P2P_1.getJson)(`http://${node.ip}:${node.port}/sync-newest-cycle`));
        if (response.newestCycle) {
            return response.newestCycle;
        }
        return null;
    };
    const newestCycle = await Utils.robustQuery(activeNodes, queryFn);
    return newestCycle.value;
}
exports.getNewestCycleFromConsensors = getNewestCycleFromConsensors;
async function getNewestCycleFromArchivers() {
    const activeArchivers = Utils.getRandomItemFromArr(State.activeArchivers, 0, 5);
    const data = {
        count: 1,
        sender: Config_1.config.ARCHIVER_PUBLIC_KEY,
    };
    Crypto.sign(data);
    const queryFn = async (node) => {
        const response = (await (0, P2P_1.postJson)(`http://${node.ip}:${node.port}/cycleinfo`, data));
        return response.cycleInfo;
    };
    const cycleInfo = await Utils.robustQuery(activeArchivers, queryFn);
    return cycleInfo.value[0];
}
exports.getNewestCycleFromArchivers = getNewestCycleFromArchivers;
async function recordArchiversReputation() {
    const activeArchivers = [...State.activeArchivers];
    const promises = activeArchivers.map((archiver) => (0, node_fetch_1.default)(`http://${archiver.ip}:${archiver.port}/cycleinfo/1`, {
        method: 'get',
        headers: { 'Content-Type': 'application/json' },
        timeout: 5000,
    }).then((res) => res.json()));
    Promise.allSettled(promises)
        .then((responses) => {
        let i = 0;
        for (const response of responses) {
            // eslint-disable-next-line security/detect-object-injection
            const archiver = activeArchivers[i];
            if (response.status === 'fulfilled') {
                const res = response.value;
                if (res && res.cycleInfo && res.cycleInfo.length > 0) {
                    const cycleRecord = res.cycleInfo[0];
                    // Set the archiver's reputation to 'up' if it is still 10 cycles behind or has higher than our current cycle
                    if (cycleRecord.counter - currentCycleCounter >= -10) {
                        State.archiversReputation.set(archiver.publicKey, 'up');
                    }
                    else {
                        Logger.mainLogger.debug(`Archiver  ${archiver.ip}:${archiver.port} has fallen behind the latest cycle`);
                        State.archiversReputation.set(archiver.publicKey, 'down');
                    }
                }
                else {
                    Logger.mainLogger.debug(`Archiver is not responding correctly ${archiver.ip}:${archiver.port}`);
                    State.archiversReputation.set(archiver.publicKey, 'down');
                }
            }
            else {
                Logger.mainLogger.debug(`Archiver is not responding ${archiver.ip}:${archiver.port}`);
                State.archiversReputation.set(archiver.publicKey, 'down');
            }
            i++;
        }
    })
        .catch((error) => {
        // Handle any errors that occurred
        console.error(error);
    });
    if (Config_1.config.VERBOSE)
        Logger.mainLogger.debug('Active archivers status', State.archiversReputation);
}
exports.recordArchiversReputation = recordArchiversReputation;
// This is called once per cycle to update to calculate the necessary shard values.
function updateShardValues(cycle) {
    const cycleShardData = {};
    // todo get current cycle..  store this by cycle?
    cycleShardData.nodeShardDataMap = new Map();
    cycleShardData.parititionShardDataMap = new Map();
    cycleShardData.nodes = NodeList.activeListByIdSorted;
    cycleShardData.cycleNumber = cycle.counter;
    cycleShardData.partitionsToSkip = new Map();
    cycleShardData.hasCompleteData = false;
    if (cycleShardData.nodes.length === 0) {
        return; // no active nodes so stop calculating values
    }
    cycleShardData.timestamp = cycle.start * 1000;
    cycleShardData.timestampEndCycle = (cycle.start + cycle.duration) * 1000;
    // save this per cycle?
    cycleShardData.shardGlobals = ShardFunctions_1.default.calculateShardGlobals(cycleShardData.nodes.length, Data_1.nodesPerConsensusGroup, Data_1.nodesPerEdge);
    if (profiler_1.profilerInstance)
        profiler_1.profilerInstance.profileSectionStart('updateShardValues_computePartitionShardDataMap1'); //13ms, #:60
    // partition shard data
    ShardFunctions_1.default.computePartitionShardDataMap(cycleShardData.shardGlobals, cycleShardData.parititionShardDataMap, 0, cycleShardData.shardGlobals.numPartitions);
    if (profiler_1.profilerInstance)
        profiler_1.profilerInstance.profileSectionEnd('updateShardValues_computePartitionShardDataMap1');
    if (profiler_1.profilerInstance)
        profiler_1.profilerInstance.profileSectionStart('updateShardValues_computePartitionShardDataMap2'); //37ms, #:60
    // generate limited data for all nodes data for all nodes.
    ShardFunctions_1.default.computeNodePartitionDataMap(cycleShardData.shardGlobals, cycleShardData.nodeShardDataMap, cycleShardData.nodes, cycleShardData.parititionShardDataMap, cycleShardData.nodes, false);
    if (profiler_1.profilerInstance)
        profiler_1.profilerInstance.profileSectionEnd('updateShardValues_computePartitionShardDataMap2');
    if (profiler_1.profilerInstance)
        profiler_1.profilerInstance.profileSectionStart('updateShardValues_computeNodePartitionData'); //22ms, #:60
    if (profiler_1.profilerInstance)
        profiler_1.profilerInstance.profileSectionEnd('updateShardValues_computeNodePartitionData');
    if (profiler_1.profilerInstance)
        profiler_1.profilerInstance.profileSectionStart('updateShardValues_computeNodePartitionDataMap2'); //232ms, #:60
    // generate lightweight data for all active nodes  (note that last parameter is false to specify the lightweight data)
    const fullDataForDebug = true; // Set this to false for performance reasons!!! setting it to true saves us from having to recalculate stuff when we dump logs.
    ShardFunctions_1.default.computeNodePartitionDataMap(cycleShardData.shardGlobals, cycleShardData.nodeShardDataMap, cycleShardData.nodes, cycleShardData.parititionShardDataMap, cycleShardData.nodes, fullDataForDebug);
    if (profiler_1.profilerInstance)
        profiler_1.profilerInstance.profileSectionEnd('updateShardValues_computeNodePartitionDataMap2');
    // console.log('cycleShardData', cycleShardData.cycleNumber)
    // console.dir(cycleShardData, { depth: null })
    const list = cycleShardData.nodes.map((n) => n['ip'] + ':' + n['port']);
    Logger.mainLogger.debug('cycleShardData', cycleShardData.cycleNumber, list.length, (0, StringifyReduce_1.stringifyReduce)(list));
    exports.shardValuesByCycle.set(cycleShardData.cycleNumber, cycleShardData);
}
const cleanShardCycleData = (cycleNumber) => {
    for (const [key] of exports.shardValuesByCycle) {
        if (key < cycleNumber) {
            exports.shardValuesByCycle.delete(key);
        }
    }
};
exports.cleanShardCycleData = cleanShardCycleData;
async function getLatestCycleRecords(count) {
    if (Config_1.config.cycleRecordsCache.enabled) {
        return await cycleDataCache.getLatestCycleRecordsFromCache(count);
    }
    const cycleInfo = await (0, cycles_1.queryLatestCycleRecords)(count);
    return Crypto.sign({ cycleInfo });
}
exports.getLatestCycleRecords = getLatestCycleRecords;
//# sourceMappingURL=Cycles.js.map