"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.isDebugMiddleware = exports.isDebugMode = void 0;
const Config_1 = require("./Config");
const Crypto = require("./Crypto");
const MAX_COUNTER_BUFFER_MILLISECONDS = 10000;
let lastCounter = 0;
function isDebugMode() {
    return !!(Config_1.config && Config_1.config.ARCHIVER_MODE && Config_1.config.ARCHIVER_MODE === 'debug');
}
exports.isDebugMode = isDebugMode;
function getDevPublicKey() {
    if (Config_1.config && Config_1.config.DevPublicKey) {
        return Config_1.config.DevPublicKey;
    }
    return '';
}
const isDebugMiddleware = (_req, res) => {
    const isDebug = isDebugMode();
    if (!isDebug) {
        try {
            //auth my by checking a signature
            if (_req.query.sig != null && _req.query.sig_counter != null) {
                const ownerPk = getDevPublicKey();
                const requestSig = _req.query.sig;
                //check if counter is valid
                const sigObj = {
                    route: _req.routerPath,
                    count: _req.query.sig_counter,
                    sign: { owner: ownerPk, sig: requestSig },
                };
                const currentCounter = parseInt(sigObj.count);
                //reguire a larger counter than before. This prevents replay attacks
                const currentTime = new Date().getTime();
                if (currentCounter > lastCounter && currentCounter <= currentTime + MAX_COUNTER_BUFFER_MILLISECONDS) {
                    const verified = Crypto.verify(sigObj);
                    if (!verified) {
                        throw new Error('FORBIDDEN. signature authentication is failed.');
                    }
                }
                else {
                    console.log(`isDebugMiddleware: currentCounter=${currentCounter}, lastCounter=${lastCounter}, currentTime=${currentTime}`);
                    throw new Error('FORBIDDEN. signature counter is failed.');
                }
                lastCounter = currentCounter; //update counter so we can't use it again
                return;
            }
            throw new Error('FORBIDDEN. Endpoint is only available in debug mode.');
        }
        catch (error) {
            // console.log(error)
            // throw new Error('FORBIDDEN. Endpoint is only available in debug mode.')
            res.code(401).send(error);
        }
    }
};
exports.isDebugMiddleware = isDebugMiddleware;
//# sourceMappingURL=DebugMode.js.map