"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getNetworkTxsListHash = exports.getTxList = exports.setTxList = exports.removeTxs = exports.addTxs = void 0;
const Logger = require("./Logger");
const StringifyReduce_1 = require("./profiler/StringifyReduce");
const crypto = require("./Crypto");
const Config_1 = require("./Config");
const fs_1 = require("fs");
const path = require("path");
const txListPath = path.join(__dirname, '..', 'tx-list-restore.json');
const rawData = (0, fs_1.readFileSync)(txListPath, 'utf8');
const ngtJson = JSON.parse(rawData);
let txList = Config_1.config.restoreNGTsFromSnapshot
    ? ngtJson
    : [];
function addTxs(addTxs) {
    try {
        for (const addTx of addTxs) {
            Logger.mainLogger.info(`Adding network tx of type ${addTx.type} and payload ${(0, StringifyReduce_1.stringifyReduce)(addTx.txData)}`);
            const { sign, ...txDataWithoutSign } = addTx.txData;
            sortedInsert(txList, {
                hash: addTx.hash,
                tx: {
                    hash: addTx.hash,
                    txData: txDataWithoutSign,
                    type: addTx.type,
                    cycle: addTx.cycle,
                    priority: addTx.priority,
                    ...(addTx.subQueueKey && { subQueueKey: addTx.subQueueKey }),
                },
            });
        }
        return true;
    }
    catch (e) {
        Logger.mainLogger.error(`ServiceQueue:addTxs: Error adding txs: ${e}`);
        return false;
    }
}
exports.addTxs = addTxs;
function removeTxs(removeTxs) {
    try {
        for (const removeTx of removeTxs) {
            const index = txList.findIndex((entry) => entry.hash === removeTx.txHash);
            if (index === -1) {
                Logger.mainLogger.error(`TxHash ${removeTx.txHash} does not exist in txList`);
            }
            else {
                txList.splice(index, 1);
            }
        }
        return true;
    }
    catch (e) {
        Logger.mainLogger.error(`ServiceQueue:removeTxs: Error removing txs: ${e}`);
        return false;
    }
}
exports.removeTxs = removeTxs;
function setTxList(_txList) {
    txList = _txList;
}
exports.setTxList = setTxList;
function getTxList() {
    return txList;
}
exports.getTxList = getTxList;
function getNetworkTxsListHash() {
    return crypto.hashObj(txList);
}
exports.getNetworkTxsListHash = getNetworkTxsListHash;
function sortedInsert(list, entry) {
    const index = list.findIndex((item) => item.tx.cycle > entry.tx.cycle ||
        (item.tx.cycle === entry.tx.cycle && item.tx.priority < entry.tx.priority) || // Compare by priority if cycle is the same
        (item.tx.cycle === entry.tx.cycle && item.tx.priority === entry.tx.priority && item.hash > entry.hash) // Compare by hash if both cycle and priority are the same
    );
    if (index === -1) {
        list.push(entry);
    }
    else {
        list.splice(index, 0, entry);
    }
}
//# sourceMappingURL=ServiceQueue.js.map