"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.calculateNetworkHash = exports.syncStateMetaData = exports.compareWithOldCyclesData = exports.syncCyclesAndNodeList = exports.buildNodeListFromStoredCycle = exports.fetchStateHashes = exports.sendToExplorer = exports.processStateMetaData = exports.subscribeMoreConsensors = exports.sendDataRequest = exports.createDataTransferConnection = exports.addDataSenders = exports.createReplaceTimeout = exports.createContactTimeout = exports.subscribeRandomNodeForDataTransfer = exports.replaceDataSender = exports.emitter = exports.dataSenders = exports.createQueryRequest = exports.createDataRequest = exports.initSocketClient = exports.unsubscribeDataSender = exports.initSocketServer = exports.currentDataSender = exports.StateMetaDataMap = exports.ArchivedCycle = exports.socketClients = exports.socketServer = void 0;
const events_1 = require("events");
const Crypto = require("../Crypto");
const NodeList = require("../NodeList");
const Storage = require("./Storage");
const Cycles = require("../Data/Cycles");
const Cycles_1 = require("../Data/Cycles");
const CycleParser_1 = require("../Data/CycleParser");
const State = require("../State");
const P2P = require("../P2P");
const Utils = require("../Utils");
const Gossip = require("./Gossip");
const util_1 = require("util");
const Config_1 = require("../Config");
const tydb_1 = require("tydb");
const lib_types_1 = require("@shardeum-foundation/lib-types");
const Logger = require("../Logger");
const nestedCounters_1 = require("../profiler/nestedCounters");
const profiler_1 = require("../profiler/profiler");
const ioclient = require("socket.io-client");
const lib_types_2 = require("@shardeum-foundation/lib-types");
let socketClient;
exports.socketClients = new Map();
const socketConnectionsTracker = new Map();
let lastSentCycleCounterToExplorer = 0;
class ArchivedCycle extends tydb_1.BaseModel {
}
exports.ArchivedCycle = ArchivedCycle;
exports.StateMetaDataMap = new Map();
exports.currentDataSender = '';
function initSocketServer(io) {
    exports.socketServer = io;
    //`socket: SocketIO.Socket` removed as an argument
    exports.socketServer.on('connection', () => {
        Logger.mainLogger.debug('Explorer has connected');
    });
}
exports.initSocketServer = initSocketServer;
function unsubscribeDataSender(publicKey) {
    Logger.mainLogger.debug('Disconnecting previous connection', publicKey);
    if (!socketClient)
        return;
    socketClient.emit('UNSUBSCRIBE', Config_1.config.ARCHIVER_PUBLIC_KEY);
    socketClient.disconnect();
    if (Config_1.config.VERBOSE)
        console.log('Killing the connection to', publicKey);
    removeDataSenders(publicKey);
    exports.currentDataSender = '';
}
exports.unsubscribeDataSender = unsubscribeDataSender;
function initSocketClient(node) {
    if (Config_1.config.VERBOSE)
        Logger.mainLogger.debug('Node Info to socker connect', node);
    if (Config_1.config.VERBOSE)
        console.log('Node Info to socker connect', node);
    try {
        const socketClient = ioclient.connect(`http://${node.ip}:${node.port}`);
        socketClient.on('connect', () => {
            Logger.mainLogger.debug('Connection to consensus node was made');
            // Send ehlo event right after connect:
            socketClient.emit('ARCHIVER_PUBLIC_KEY', Config_1.config.ARCHIVER_PUBLIC_KEY);
        });
        socketClient.once('disconnect', async () => {
            Logger.mainLogger.debug(`Connection request is refused by the consensor node ${node.ip}:${node.port}`);
            console.log(`Connection request is refused by the consensor node ${node.ip}:${node.port}`);
            // socketClients.delete(node.publicKey)
            // await Utils.sleep(3000)
            // if (socketClients.has(node.publicKey)) replaceDataSender(node.publicKey)
            socketConnectionsTracker.set(node.publicKey, 'disconnected');
        });
        socketClient.on('DATA', (data) => {
            const newData = lib_types_2.Utils.safeJsonParse(data);
            if (!newData || !newData.responses)
                return;
            if (newData.recipient !== State.getNodeInfo().publicKey) {
                Logger.mainLogger.debug('This data is not meant for this archiver');
                return;
            }
            // If tag is invalid, dont keepAlive, END
            if (Crypto.authenticate(newData) === false) {
                Logger.mainLogger.debug('This data cannot be authenticated');
                console.log('Unsubscribe 1', node.publicKey);
                unsubscribeDataSender(node.publicKey);
                return;
            }
            if (newData.responses.STATE_METADATA.length > 0)
                Logger.mainLogger.debug('New DATA', newData.responses);
            else
                Logger.mainLogger.debug('State metadata is empty');
            exports.currentDataSender = newData.publicKey;
            if (newData.responses && newData.responses.STATE_METADATA) {
                // Logger.mainLogger.debug('New DATA from consensor STATE_METADATA', newData.publicKey, newData.responses.STATE_METADATA)
                // let hashArray: any = Gossip.convertStateMetadataToHashArray(newData.responses.STATE_METADATA[0])
                for (const stateMetadata of newData.responses.STATE_METADATA) {
                    exports.StateMetaDataMap.set(stateMetadata.counter, stateMetadata);
                    Gossip.sendGossip('hashes', stateMetadata);
                }
            }
            exports.socketServer.emit('DATA', newData);
            const sender = exports.dataSenders.get(newData.publicKey);
            // If publicKey is not in dataSenders, dont keepAlive, END
            if (!sender) {
                Logger.mainLogger.debug('NO SENDER');
                return;
            }
            // If unexpected data type from sender, dont keepAlive, END
            const newDataTypes = Object.keys(newData.responses);
            for (const type of newDataTypes) {
                if (sender.types.includes(type) === false) {
                    Logger.mainLogger.debug(`NEW DATA type ${type} not included in sender's types: ${lib_types_2.Utils.safeStringify(sender.types)}`);
                    return;
                }
            }
            setImmediate(processData, newData);
        });
    }
    catch (error) {
        console.error('Error occurred during socket connection:', error);
    }
}
exports.initSocketClient = initSocketClient;
function createDataRequest(type, lastData, recipientPk) {
    return Crypto.tag({
        type,
        lastData,
    }, recipientPk);
}
exports.createDataRequest = createDataRequest;
function createQueryRequest(type, lastData, recipientPk) {
    return Crypto.tag({
        type,
        lastData,
    }, recipientPk);
}
exports.createQueryRequest = createQueryRequest;
exports.dataSenders = new Map();
const timeoutPadding = 1000;
exports.emitter = new events_1.EventEmitter();
async function replaceDataSender(publicKey) {
    nestedCounters_1.nestedCountersInstance.countEvent('archiver', 'replace_data_sender');
    if (NodeList.getActiveNodeCount() < 2) {
        Logger.mainLogger.debug('There is only one active node in the network. Unable to replace data sender');
        const sender = exports.dataSenders.get(publicKey);
        if (sender && sender.replaceTimeout) {
            nestedCounters_1.nestedCountersInstance.countEvent('archiver', 'clear_replace_timeout');
            clearTimeout(sender.replaceTimeout);
            sender.replaceTimeout = null;
            sender.replaceTimeout = createReplaceTimeout(publicKey);
        }
        return;
    }
    Logger.mainLogger.debug(`replaceDataSender: replacing ${publicKey}`);
    // Remove old dataSender
    const removedSenders = removeDataSenders(publicKey);
    if (removedSenders.length < 1) {
        Logger.mainLogger.debug('replaceDataSender failed: old sender not removed');
    }
    // Pick a new dataSender
    const newSenderInfo = selectNewDataSender(publicKey);
    if (!newSenderInfo) {
        Logger.mainLogger.error('Unable to select a new data sender.');
        return;
    }
    Logger.mainLogger.debug('Before sleep', publicKey, newSenderInfo);
    await Utils.sleep(1000); // Wait about 1s to be sure that socket client connection is killed
    initSocketClient(newSenderInfo);
    const newSender = {
        nodeInfo: newSenderInfo,
        types: [lib_types_1.P2P.SnapshotTypes.TypeNames.CYCLE, lib_types_1.P2P.SnapshotTypes.TypeNames.STATE_METADATA],
        contactTimeout: createContactTimeout(newSenderInfo.publicKey, 'This timeout is created during newSender selection', 2 * Cycles_1.currentCycleDuration),
        replaceTimeout: createReplaceTimeout(newSenderInfo.publicKey),
    };
    // Add new dataSender to dataSenders
    addDataSenders(newSender);
    Logger.mainLogger.debug(`replaceDataSender: added new sender ${newSenderInfo.publicKey} to dataSenders`);
    // Send dataRequest to new dataSender
    const dataRequest = {
        dataRequestCycle: createDataRequest(lib_types_1.P2P.SnapshotTypes.TypeNames.CYCLE, Cycles.getCurrentCycleCounter(), publicKey),
        dataRequestStateMetaData: createDataRequest(lib_types_1.P2P.SnapshotTypes.TypeNames.STATE_METADATA, Cycles_1.lastProcessedMetaData, publicKey),
        publicKey: State.getNodeInfo().publicKey,
        nodeInfo: State.getNodeInfo(),
    };
    sendDataRequest(newSender, dataRequest);
}
exports.replaceDataSender = replaceDataSender;
async function subscribeRandomNodeForDataTransfer() {
    let retry = 0;
    let nodeSubscribedFail = true;
    // Set randomly select a consensor as dataSender
    while (nodeSubscribedFail && retry < 10) {
        const randomConsensor = NodeList.getRandomActiveNodes()[0];
        const connectionStatus = await createDataTransferConnection(randomConsensor);
        if (connectionStatus)
            nodeSubscribedFail = false;
        else
            retry++;
    }
    if (nodeSubscribedFail) {
        Logger.mainLogger.error('The archiver fails to subscribe to any node for data transfer! and exit the network.');
        await State.exitArchiver();
    }
}
exports.subscribeRandomNodeForDataTransfer = subscribeRandomNodeForDataTransfer;
/**
 * Sets timeout to current cycle duration + some padding
 * Removes sender from dataSenders on timeout
 * Select a new dataSender
 */
function createContactTimeout(publicKey, msg = '', timeout = null) {
    // TODO: check what is the best contact timeout
    let ms;
    if (timeout)
        ms = timeout;
    else if (Cycles_1.currentCycleDuration > 0)
        ms = 1.5 * Cycles_1.currentCycleDuration + timeoutPadding;
    else
        ms = 1.5 * 60 * 1000 + timeoutPadding;
    if (Config_1.config.experimentalSnapshot)
        ms = 15 * 1000; // Change contact timeout to 15s for now
    Logger.mainLogger.debug('Created contact timeout: ' + ms);
    nestedCounters_1.nestedCountersInstance.countEvent('archiver', 'contact_timeout_created');
    return setTimeout(() => {
        // Logger.mainLogger.debug('nestedCountersInstance', nestedCountersInstance)
        if (nestedCounters_1.nestedCountersInstance)
            nestedCounters_1.nestedCountersInstance.countEvent('archiver', 'contact_timeout');
        Logger.mainLogger.debug('REPLACING sender due to CONTACT timeout', msg);
        console.log('REPLACING sender due to CONTACT timeout', msg, publicKey);
        replaceDataSender(publicKey);
    }, ms);
}
exports.createContactTimeout = createContactTimeout;
function createReplaceTimeout(publicKey) {
    const ms = Config_1.config.DATASENDER_TIMEOUT || 1000 * 60 * 60;
    return setTimeout(() => {
        nestedCounters_1.nestedCountersInstance.countEvent('archiver', 'replace_timeout');
        Logger.mainLogger.debug('ROTATING sender due to ROTATION timeout');
        replaceDataSender(publicKey);
    }, ms);
}
exports.createReplaceTimeout = createReplaceTimeout;
function addDataSenders(...senders) {
    for (const sender of senders) {
        exports.dataSenders.set(sender.nodeInfo.publicKey, sender);
        exports.currentDataSender = sender.nodeInfo.publicKey;
    }
}
exports.addDataSenders = addDataSenders;
function removeDataSenders(publicKey) {
    Logger.mainLogger.debug(`${new Date()}: Removing data sender ${publicKey}`);
    const removedSenders = [];
    // console.log('removeDataSenders', dataSenders)
    // Logger.mainLogger.debug('removeDataSenders', dataSenders)
    for (const [key, sender] of exports.dataSenders) {
        // if (config.VERBOSE) Logger.mainLogger.debug(publicKey, key)
        if (key === publicKey && sender) {
            // Clear contactTimeout associated with this sender
            if (sender.contactTimeout) {
                clearTimeout(sender.contactTimeout);
                sender.contactTimeout = null;
            }
            if (sender.replaceTimeout) {
                clearTimeout(sender.replaceTimeout);
                sender.replaceTimeout = null;
            }
            nestedCounters_1.nestedCountersInstance.countEvent('archiver', 'remove_data_sender');
            // Record which sender was removed
            removedSenders.push(sender);
            // Delete sender from dataSenders
            exports.dataSenders.delete(key);
        }
    }
    return removedSenders;
}
function selectNewDataSender(publicKey) {
    // Randomly pick an active node
    const activeList = NodeList.getActiveList();
    const newSender = activeList[Math.floor(Math.random() * activeList.length)];
    Logger.mainLogger.debug('New data sender is selected', newSender);
    if (newSender) {
        unsubscribeDataSender(publicKey);
        // initSocketClient(newSender)
    }
    return newSender;
}
async function createDataTransferConnection(newSenderInfo) {
    initSocketClient(newSenderInfo);
    let count = 0;
    while (!exports.socketClients.has(newSenderInfo.publicKey) && count <= 50) {
        await Utils.sleep(100);
        count++;
        if (count === 50) {
            // This means the socket connection to the node is not successful.
            return false;
        }
    }
    if (socketConnectionsTracker.get(newSenderInfo.publicKey) === 'disconnected')
        return false;
    await Utils.sleep(200); // Wait 1s before sending data request to be sure if the connection is not refused
    if (socketConnectionsTracker.get(newSenderInfo.publicKey) === 'disconnected')
        return false;
    const newSender = {
        nodeInfo: newSenderInfo,
        types: [lib_types_1.P2P.SnapshotTypes.TypeNames.CYCLE, lib_types_1.P2P.SnapshotTypes.TypeNames.STATE_METADATA],
        contactTimeout: createContactTimeout(newSenderInfo.publicKey, 'This timeout is created during newSender selection', 2 * Cycles_1.currentCycleDuration),
        replaceTimeout: createReplaceTimeout(newSenderInfo.publicKey),
    };
    // Add new dataSender to dataSenders
    addDataSenders(newSender);
    Logger.mainLogger.debug(`replaceDataSender: added new sender ${newSenderInfo.publicKey} to dataSenders`);
    // Send dataRequest to new dataSender
    const dataRequest = {
        dataRequestCycle: createDataRequest(lib_types_1.P2P.SnapshotTypes.TypeNames.CYCLE, Cycles.getCurrentCycleCounter(), State.getNodeInfo().publicKey),
        dataRequestStateMetaData: createDataRequest(lib_types_1.P2P.SnapshotTypes.TypeNames.STATE_METADATA, Cycles_1.lastProcessedMetaData, State.getNodeInfo().publicKey),
        publicKey: State.getNodeInfo().publicKey,
        nodeInfo: State.getNodeInfo(),
    };
    sendDataRequest(newSender, dataRequest);
    return true;
}
exports.createDataTransferConnection = createDataTransferConnection;
function sendDataRequest(sender, dataRequest) {
    const taggedDataRequest = Crypto.tag(dataRequest, sender.nodeInfo.publicKey);
    Logger.mainLogger.info('Sending tagged data request to consensor.', sender);
    if (exports.socketClients.has(sender.nodeInfo.publicKey))
        exports.emitter.emit('selectNewDataSender', sender.nodeInfo, taggedDataRequest);
}
exports.sendDataRequest = sendDataRequest;
async function subscribeMoreConsensors(numbersToSubscribe) {
    Logger.mainLogger.info(`Subscribing ${numbersToSubscribe} more consensors for tagged data request`);
    console.log('numbersToSubscribe', numbersToSubscribe);
    for (let i = 0; i < numbersToSubscribe; i++) {
        await Utils.sleep(60000);
        // Pick a new dataSender
        const activeList = NodeList.getActiveList();
        const newSenderInfo = activeList[Math.floor(Math.random() * activeList.length)];
        Logger.mainLogger.debug('New data sender is selected', newSenderInfo);
        if (!newSenderInfo) {
            Logger.mainLogger.error('Unable to select a new data sender.');
            continue;
        }
        initSocketClient(newSenderInfo);
        const newSender = {
            nodeInfo: newSenderInfo,
            types: [lib_types_1.P2P.SnapshotTypes.TypeNames.CYCLE, lib_types_1.P2P.SnapshotTypes.TypeNames.STATE_METADATA],
            contactTimeout: createContactTimeout(newSenderInfo.publicKey, 'This timeout is created during newSender selection', 2 * Cycles_1.currentCycleDuration),
            replaceTimeout: createReplaceTimeout(newSenderInfo.publicKey),
        };
        // Add new dataSender to dataSenders
        addDataSenders(newSender);
        Logger.mainLogger.debug(`replaceDataSender: added new sender ${newSenderInfo.publicKey} to dataSenders`);
        // Send dataRequest to new dataSender
        const dataRequest = {
            dataRequestCycle: createDataRequest(lib_types_1.P2P.SnapshotTypes.TypeNames.CYCLE, Cycles.getCurrentCycleCounter(), State.getNodeInfo().publicKey),
            dataRequestStateMetaData: createDataRequest(lib_types_1.P2P.SnapshotTypes.TypeNames.STATE_METADATA, Cycles_1.lastProcessedMetaData, State.getNodeInfo().publicKey),
            publicKey: State.getNodeInfo().publicKey,
            nodeInfo: State.getNodeInfo(),
        };
        sendDataRequest(newSender, dataRequest);
    }
}
exports.subscribeMoreConsensors = subscribeMoreConsensors;
async function sendDataQuery(consensorNode, dataQuery, validateFn) {
    const taggedDataQuery = Crypto.tag(dataQuery, consensorNode.publicKey);
    const result = await queryDataFromNode(consensorNode, taggedDataQuery, validateFn);
    return result;
}
async function processData(newData) {
    // Get sender entry
    const sender = exports.dataSenders.get(newData.publicKey);
    // If no sender entry, remove publicKey from senders, END
    if (!sender) {
        Logger.mainLogger.error('No sender found for this data');
        return;
    }
    if (sender.nodeInfo.publicKey !== exports.currentDataSender) {
        Logger.mainLogger.error(`Sender ${sender.nodeInfo.publicKey} is not current data sender.`);
    }
    // Clear senders contactTimeout, if it has one
    if (sender.contactTimeout) {
        Logger.mainLogger.debug('Clearing contact timeout.');
        clearTimeout(sender.contactTimeout);
        sender.contactTimeout = null;
        nestedCounters_1.nestedCountersInstance.countEvent('archiver', 'clear_contact_timeout');
    }
    const newDataTypes = Object.keys(newData.responses);
    for (const type of newDataTypes) {
        // Process data depending on type
        switch (type) {
            case lib_types_1.P2P.SnapshotTypes.TypeNames.CYCLE: {
                Logger.mainLogger.debug('Processing CYCLE data');
                (0, Cycles_1.processCycles)(newData.responses.CYCLE);
                // socketServer.emit('ARCHIVED_CYCLE', 'CYCLE')
                if (newData.responses.CYCLE.length > 0) {
                    for (const cycle of newData.responses.CYCLE) {
                        const archivedCycle = {
                            cycleRecord: cycle,
                            cycleMarker: cycle.marker,
                            data: undefined,
                            receipt: undefined,
                            summary: undefined,
                            _id: undefined,
                        };
                        archivedCycle.cycleRecord = cycle;
                        archivedCycle.cycleMarker = cycle.marker;
                        Cycles.CycleChain.set(cycle.counter, cycle);
                        await Storage.insertArchivedCycle(archivedCycle);
                    }
                }
                else {
                    Logger.mainLogger.error('Received empty newData.responses.CYCLE', newData.responses);
                }
                break;
            }
            case lib_types_1.P2P.SnapshotTypes.TypeNames.STATE_METADATA: {
                Logger.mainLogger.debug('Processing STATE_METADATA');
                processStateMetaData(newData.responses);
                break;
            }
            default: {
                // If data type not recognized, remove sender from dataSenders
                Logger.mainLogger.error('Unknown data type detected', type);
                removeDataSenders(newData.publicKey);
            }
        }
    }
    // Set new contactTimeout for sender. Postpone sender removal because data is still received from consensor
    if (Cycles_1.currentCycleDuration > 0) {
        nestedCounters_1.nestedCountersInstance.countEvent('archiver', 'postpone_contact_timeout');
        sender.contactTimeout = createContactTimeout(sender.nodeInfo.publicKey, 'This timeout is created after processing data');
    }
}
async function processStateMetaData(response) {
    Logger.mainLogger.error('response', response);
    const STATE_METADATA = response.STATE_METADATA;
    if (!STATE_METADATA || STATE_METADATA.length === 0) {
        Logger.mainLogger.error('Invalid STATE_METADATA provided to processStateMetaData function', STATE_METADATA);
        return;
    }
    profiler_1.profilerInstance.profileSectionStart('state_metadata');
    for (const stateMetaData of STATE_METADATA) {
        let data;
        let receipt;
        let summary;
        // [TODO] validate the state data by robust querying other nodes
        // store state hashes to archivedCycle
        for (const stateHashesForCycle of stateMetaData.stateHashes) {
            const parentCycle = Cycles.CycleChain.get(stateHashesForCycle.counter);
            if (!parentCycle) {
                Logger.mainLogger.error('Unable to find parent cycle for cycle', stateHashesForCycle.counter);
                continue;
            }
            data = {
                parentCycle: parentCycle ? parentCycle.marker : '',
                networkHash: stateHashesForCycle.networkHash,
                partitionHashes: stateHashesForCycle.partitionHashes,
            };
            await Storage.updateArchivedCycle(data.parentCycle, 'data', data);
            // if (!processedCounters[parentCycle.counter]) {
            //   processedCounters[parentCycle.counter] = true
            // }
            // Cycles.setLastProcessedMetaDataCounter(parentCycle.counter)
        }
        // store receipt hashes to archivedCycle
        for (const receiptHashesForCycle of stateMetaData.receiptHashes) {
            const parentCycle = Cycles.CycleChain.get(receiptHashesForCycle.counter);
            if (!parentCycle) {
                Logger.mainLogger.error('Unable to find parent cycle for cycle', receiptHashesForCycle.counter);
                continue;
            }
            receipt = {
                parentCycle: parentCycle ? parentCycle.marker : '',
                networkHash: receiptHashesForCycle.networkReceiptHash,
                partitionHashes: receiptHashesForCycle.receiptMapHashes,
                partitionMaps: {},
                partitionTxs: {},
            };
            await Storage.updateArchivedCycle(receipt.parentCycle, 'receipt', receipt);
            // Cycles.setLastProcessedMetaDataCounter(parentCycle.counter)
            // Query receipt maps from other nodes and store it
            if (receiptHashesForCycle.receiptMapHashes) {
                let isDownloadSuccess = false;
                let retry = 0;
                let sleepCount = 0;
                const failedPartitions = new Map();
                const coveredPartitions = new Map();
                const downloadedReceiptMaps = new Map();
                const shouldProcessReceipt = (cycle, partition) => {
                    if (cycle === receiptHashesForCycle.counter - 1)
                        if (failedPartitions.has(partition) || !coveredPartitions.has(partition))
                            return true;
                    return false;
                };
                const cycleActiveNodesSize = parentCycle.active + parentCycle.activated.length - parentCycle.removed.length;
                while (!isDownloadSuccess && sleepCount < 20) {
                    const randomConsensor = NodeList.getRandomActiveNodes()[0];
                    const queryRequest = createQueryRequest('RECEIPT_MAP', receiptHashesForCycle.counter - 1, randomConsensor.publicKey);
                    const { success, completed, failed, covered, blobs } = (await sendDataQuery(randomConsensor, queryRequest, shouldProcessReceipt));
                    if (success) {
                        for (const partition of failed) {
                            failedPartitions.set(partition, true);
                        }
                        for (const partition of completed) {
                            if (failedPartitions.has(partition))
                                failedPartitions.delete(partition);
                        }
                        for (const partition of covered) {
                            coveredPartitions.set(partition, true);
                        }
                        for (const partition in blobs) {
                            // eslint-disable-next-line security/detect-object-injection
                            downloadedReceiptMaps.set(partition, blobs[partition]);
                        }
                    }
                    isDownloadSuccess = failedPartitions.size === 0 && coveredPartitions.size === cycleActiveNodesSize;
                    if (isDownloadSuccess) {
                        Logger.mainLogger.debug('Data query for receipt map is completed for cycle', parentCycle.counter);
                        Logger.mainLogger.debug('Total downloaded receipts', downloadedReceiptMaps.size);
                        // if (!processedCounters[receiptHashesForCycle.counter - 1]) {
                        //   processedCounters[receiptHashesForCycle.counter - 1] = true
                        // }
                        sendToExplorer(receiptHashesForCycle.counter - 1);
                        // let receiptMapsToForward = []
                        // for (let [partition, receiptMap] of downloadedReceiptMaps) {
                        //   receiptMapsToForward.push(receiptMap)
                        // }
                        // receiptMapsToForward = receiptMapsToForward.filter(
                        //   (receipt) => receipt.cycle === parentCycle.counter
                        // )
                        // Logger.mainLogger.debug(
                        //   'receiptMapsToForward',
                        //   receiptMapsToForward.length
                        // )
                        // socketServer.emit('RECEIPT_MAP', receiptMapsToForward)
                        break;
                    }
                    retry += 1;
                    if (!isDownloadSuccess && retry >= NodeList.getActiveNodeCount()) {
                        Logger.mainLogger.debug('Sleeping for 5 sec before retrying download again for cycle', parentCycle.counter);
                        await Utils.sleep(5000);
                        retry = 0;
                        sleepCount += 1;
                    }
                }
                if (!isDownloadSuccess) {
                    Logger.mainLogger.debug(`Downloading receipt map for cycle ${parentCycle.counter} has failed.`);
                    Logger.mainLogger.debug(`There are ${failedPartitions.size} failed partitions in cycle ${parentCycle.counter}.`);
                }
            }
        }
        // store summary hashes to archivedCycle
        for (const summaryHashesForCycle of stateMetaData.summaryHashes) {
            const parentCycle = Cycles.CycleChain.get(summaryHashesForCycle.counter);
            if (!parentCycle) {
                Logger.mainLogger.error('Unable to find parent cycle for cycle', summaryHashesForCycle.counter);
                continue;
            }
            summary = {
                parentCycle: parentCycle ? parentCycle.marker : '',
                networkHash: summaryHashesForCycle.networkSummaryHash,
                partitionHashes: summaryHashesForCycle.summaryHashes,
                partitionBlobs: {},
            };
            await Storage.updateArchivedCycle(summary.parentCycle, 'summary', summary);
            Cycles.setLastProcessedMetaDataCounter(parentCycle.counter);
            // // Query summary blobs from other nodes and store it
            // if (summaryHashesForCycle.summaryHashes) {
            //   let isDownloadSuccess = false
            //   let retry = 0
            //   let sleepCount = 0
            //   let failedPartitions = new Map()
            //   let coveredPartitions = new Map()
            //   let downloadedBlobs = new Map()
            //   let shouldProcessBlob = (cycle: number, partition: number) => {
            //     if (cycle === summaryHashesForCycle.counter - 1)
            //       if (
            //         failedPartitions.has(partition) ||
            //         !coveredPartitions.has(partition)
            //       )
            //         return true
            //     return false
            //   }
            //   while (!isDownloadSuccess && sleepCount < 20) {
            //     let randomConsensor = NodeList.getRandomActiveNode()
            //     const queryRequest = createQueryRequest(
            //       'SUMMARY_BLOB',
            //       summaryHashesForCycle.counter - 1,
            //       randomConsensor.publicKey
            //     )
            //     let { success, completed, failed, covered, blobs } =
            //       await sendDataQuery(
            //         randomConsensor,
            //         queryRequest,
            //         shouldProcessBlob
            //       )
            //     if (success) {
            //       for (let partition of failed) {
            //         failedPartitions.set(partition, true)
            //       }
            //       for (let partition of completed) {
            //         if (failedPartitions.has(partition))
            //           failedPartitions.delete(partition)
            //       }
            //       for (let partition of covered) {
            //         coveredPartitions.set(partition, true)
            //       }
            //       for (let partition in blobs) {
            //         downloadedBlobs.set(partition, blobs[partition])
            //       }
            //     }
            //     isDownloadSuccess =
            //       failedPartitions.size === 0 && coveredPartitions.size === 4096
            //     if (isDownloadSuccess) {
            //       Logger.mainLogger.debug(
            //         'Data query for summary blob is completed for cycle',
            //         parentCycle.counter
            //       )
            //       Logger.mainLogger.debug(
            //         'Total downloaded blobs',
            //         downloadedBlobs.size
            //       )
            //       break
            //     }
            //     retry += 1
            //     if (!isDownloadSuccess && retry >= NodeList.getActiveNodeCount()) {
            //       Logger.mainLogger.debug(
            //         'Sleeping for 5 sec before retrying download again for cycle',
            //         parentCycle.counter
            //       )
            //       await Utils.sleep(5000)
            //       retry = 0
            //       sleepCount += 1
            //     }
            //   }
            //   if (!isDownloadSuccess) {
            //     Logger.mainLogger.debug(
            //       `Downloading summary blob for cycle ${parentCycle.counter} has failed.`
            //     )
            //     Logger.mainLogger.debug(
            //       `There are ${failedPartitions.size} failed partitions in cycle ${parentCycle.counter}.`
            //     )
            //   }
            // }
            // if (!processedCounters[parentCycle.counter]) {
            //   processedCounters[parentCycle.counter] = true
            // }
        }
    }
    // if (socketServer)  {
    //   Logger.mainLogger.debug('Finished processing state metadata...', processedCounters, lastSentCycleCounterToExplorer)
    //   for (let counter of Object.keys(processedCounters)) {
    //     let start = parseInt(counter)
    //     let end = parseInt(counter)
    //     if (lastSentCycleCounterToExplorer === 0 || parseInt(counter) > lastSentCycleCounterToExplorer + 1) {
    //       start = lastSentCycleCounterToExplorer + 1
    //     } else {
    //       continue
    //     }
    //     Logger.mainLogger.debug('start, end', start, end)
    //     Logger.mainLogger.debug(start,end)
    //     let completedArchivedCycle = await Storage.queryAllArchivedCyclesBetween(start, end)
    //     Logger.mainLogger.debug('completedArchivedCycle', completedArchivedCycle.length, completedArchivedCycle)
    //     let signedDataToSend = Crypto.sign({
    //       archivedCycles: completedArchivedCycle,
    //     })
    //     Logger.mainLogger.debug('Sending completed archived_cycle to explorer', signedDataToSend)
    //     lastSentCycleCounterToExplorer = end
    //     if (socketServer) socketServer.emit('ARCHIVED_CYCLE', signedDataToSend)
    //   }
    // }
    profiler_1.profilerInstance.profileSectionEnd('state_metadata');
}
exports.processStateMetaData = processStateMetaData;
async function sendToExplorer(counter) {
    if (exports.socketServer) {
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        let completedArchivedCycle;
        if (lastSentCycleCounterToExplorer === 0) {
            completedArchivedCycle = await Storage.queryAllArchivedCyclesBetween(lastSentCycleCounterToExplorer, counter);
            Logger.mainLogger.debug('start, end', lastSentCycleCounterToExplorer, counter);
        }
        else {
            completedArchivedCycle = await Storage.queryAllArchivedCyclesBetween(counter, counter);
            Logger.mainLogger.debug('start, end', counter, counter);
        }
        const signedDataToSend = Crypto.sign({
            archivedCycles: completedArchivedCycle,
        });
        Logger.mainLogger.debug('Sending completed archived_cycle to explorer', signedDataToSend);
        lastSentCycleCounterToExplorer = counter;
        if (exports.socketServer)
            exports.socketServer.emit('ARCHIVED_CYCLE', signedDataToSend);
    }
}
exports.sendToExplorer = sendToExplorer;
async function fetchStateHashes(archivers) {
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    function _isSameStateHashes(info1, info2) {
        const cm1 = Utils.deepCopy(info1);
        const cm2 = Utils.deepCopy(info2);
        delete cm1.currentTime;
        delete cm2.currentTime;
        const equivalent = (0, util_1.isDeepStrictEqual)(cm1, cm2);
        return equivalent;
    }
    const queryFn = async (node) => {
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        const response = await P2P.getJson(`http://${node.ip}:${node.port}/statehashes`);
        return response.stateHashes;
    };
    const stateHashes = await Utils.robustQuery(archivers, queryFn, _isSameStateHashes);
    return stateHashes.value[0];
}
exports.fetchStateHashes = fetchStateHashes;
async function buildNodeListFromStoredCycle(lastStoredCycle) {
    Logger.mainLogger.debug('lastStoredCycle', lastStoredCycle);
    Logger.mainLogger.debug(`Syncing till cycle ${lastStoredCycle.counter}...`);
    const cyclesToGet = 2 * Math.floor(Math.sqrt(lastStoredCycle.active)) + 2;
    Logger.mainLogger.debug(`Cycles to get is ${cyclesToGet}`);
    const CycleChain = [];
    const squasher = new CycleParser_1.ChangeSquasher();
    CycleChain.unshift(lastStoredCycle);
    squasher.addChange((0, CycleParser_1.parse)(CycleChain[0]));
    do {
        // Get prevCycles from the network
        let end = CycleChain[0].counter - 1;
        let start = end - cyclesToGet;
        if (start < 0)
            start = 0;
        if (end < start)
            end = start;
        Logger.mainLogger.debug(`Getting cycles ${start} - ${end}...`);
        const prevCycles = await Storage.queryCycleRecordsBetween(start, end);
        // If prevCycles is empty, start over
        if (prevCycles.length < 1)
            throw new Error('Got empty previous cycles');
        prevCycles.sort((a, b) => (a.counter > b.counter ? -1 : 1));
        // Add prevCycles to our cycle chain
        let prepended = 0;
        for (const prevCycle of prevCycles) {
            // Prepend the cycle to our cycle chain
            CycleChain.unshift(prevCycle);
            squasher.addChange((0, CycleParser_1.parse)(prevCycle));
            prepended++;
            if (squasher.final.updated.length >= (0, CycleParser_1.activeNodeCount)(lastStoredCycle) &&
                squasher.final.added.length >= (0, CycleParser_1.totalNodeCount)(lastStoredCycle)) {
                break;
            }
        }
        Logger.mainLogger.debug(`Got ${squasher.final.updated.length} active nodes, need ${(0, CycleParser_1.activeNodeCount)(lastStoredCycle)}`);
        Logger.mainLogger.debug(`Got ${squasher.final.added.length} total nodes, need ${(0, CycleParser_1.totalNodeCount)(lastStoredCycle)}`);
        if (squasher.final.added.length < (0, CycleParser_1.totalNodeCount)(lastStoredCycle))
            Logger.mainLogger.debug('Short on nodes. Need to get more cycles. Cycle:' + lastStoredCycle.counter);
        // If you weren't able to prepend any of the prevCycles, start over
        if (prepended < 1)
            throw new Error('Unable to prepend any previous cycles');
    } while (squasher.final.updated.length < (0, CycleParser_1.activeNodeCount)(lastStoredCycle) ||
        squasher.final.added.length < (0, CycleParser_1.totalNodeCount)(lastStoredCycle));
    (0, CycleParser_1.applyNodeListChange)(squasher.final);
    Logger.mainLogger.debug('NodeList after sync', NodeList.getActiveList());
    Cycles.setCurrentCycleCounter(lastStoredCycle.counter);
    Cycles.setCurrentCycleMarker(lastStoredCycle.marker);
    Cycles.setCurrentCycleDuration(lastStoredCycle.duration);
    Logger.mainLogger.debug('Latest cycle after sync', lastStoredCycle.counter);
}
exports.buildNodeListFromStoredCycle = buildNodeListFromStoredCycle;
async function syncCyclesAndNodeList(lastStoredCycleCount = 0) {
    // Get the networks newest cycle as the anchor point for sync
    Logger.mainLogger.debug('Getting newest cycle...');
    const cycleToSyncTo = await (0, Cycles_1.getNewestCycleFromArchivers)();
    Logger.mainLogger.debug('cycleToSyncTo', cycleToSyncTo);
    Logger.mainLogger.debug(`Syncing till cycle ${cycleToSyncTo.counter}...`);
    const cyclesToGet = 2 * Math.floor(Math.sqrt(cycleToSyncTo.active)) + 2;
    Logger.mainLogger.debug(`Cycles to get is ${cyclesToGet}`);
    const CycleChain = [];
    const squasher = new CycleParser_1.ChangeSquasher();
    CycleChain.unshift(cycleToSyncTo);
    squasher.addChange((0, CycleParser_1.parse)(CycleChain[0]));
    do {
        // Get prevCycles from the network
        let end = CycleChain[0].counter - 1;
        let start = end - cyclesToGet;
        if (start < 0)
            start = 0;
        if (end < start)
            end = start;
        Logger.mainLogger.debug(`Getting cycles ${start} - ${end}...`);
        const prevCycles = await (0, Cycles_1.fetchCycleRecords)(start, end);
        // If prevCycles is empty, start over
        if (prevCycles.length < 1)
            throw new Error('Got empty previous cycles');
        prevCycles.sort((a, b) => (a.counter > b.counter ? -1 : 1));
        // Add prevCycles to our cycle chain
        let prepended = 0;
        for (const prevCycle of prevCycles) {
            // Stop prepending prevCycles if one of them is invalid
            if ((0, Cycles_1.validateCycle)(prevCycle, CycleChain[0]) === false) {
                Logger.mainLogger.error(`Record ${prevCycle.counter} failed validation`);
                break;
            }
            // Prepend the cycle to our cycle chain
            CycleChain.unshift(prevCycle);
            squasher.addChange((0, CycleParser_1.parse)(prevCycle));
            prepended++;
            if (squasher.final.updated.length >= (0, CycleParser_1.activeNodeCount)(cycleToSyncTo) &&
                squasher.final.added.length >= (0, CycleParser_1.totalNodeCount)(cycleToSyncTo)) {
                break;
            }
        }
        Logger.mainLogger.debug(`Got ${squasher.final.updated.length} active nodes, need ${(0, CycleParser_1.activeNodeCount)(cycleToSyncTo)}`);
        Logger.mainLogger.debug(`Got ${squasher.final.added.length} total nodes, need ${(0, CycleParser_1.totalNodeCount)(cycleToSyncTo)}`);
        if (squasher.final.added.length < (0, CycleParser_1.totalNodeCount)(cycleToSyncTo))
            Logger.mainLogger.debug('Short on nodes. Need to get more cycles. Cycle:' + cycleToSyncTo.counter);
        // If you weren't able to prepend any of the prevCycles, start over
        if (prepended < 1)
            throw new Error('Unable to prepend any previous cycles');
    } while (squasher.final.updated.length < (0, CycleParser_1.activeNodeCount)(cycleToSyncTo) ||
        squasher.final.added.length < (0, CycleParser_1.totalNodeCount)(cycleToSyncTo));
    (0, CycleParser_1.applyNodeListChange)(squasher.final);
    Logger.mainLogger.debug('NodeList after sync', NodeList.getActiveList());
    for (let i = 0; i < CycleChain.length; i++) {
        // eslint-disable-next-line security/detect-object-injection
        const record = CycleChain[i];
        Cycles.CycleChain.set(record.counter, { ...record });
        Logger.mainLogger.debug('Inserting archived cycle for counter', record.counter);
        const archivedCycle = createArchivedCycle(record);
        await Storage.insertArchivedCycle(archivedCycle);
        Cycles.setCurrentCycleCounter(record.counter);
        Cycles.setCurrentCycleMarker(record.marker);
    }
    Logger.mainLogger.debug('Cycle chain is synced. Size of CycleChain', Cycles.CycleChain.size);
    // Download old cycle Records
    let endCycle = CycleChain[0].counter - 1;
    Logger.mainLogger.debug('endCycle counter', endCycle, 'lastStoredCycleCount', lastStoredCycleCount);
    if (endCycle > lastStoredCycleCount) {
        Logger.mainLogger.debug(`Downloading old cycles from cycles ${lastStoredCycleCount} to cycle ${endCycle}!`);
    }
    let savedCycleRecord = CycleChain[0];
    while (endCycle > lastStoredCycleCount) {
        let nextEnd = endCycle - 10000; // Downloading max 1000 cycles each time
        if (nextEnd < 0)
            nextEnd = 0;
        Logger.mainLogger.debug(`Getting cycles ${nextEnd} - ${endCycle} ...`);
        const prevCycles = await (0, Cycles_1.fetchCycleRecords)(nextEnd, endCycle);
        // If prevCycles is empty, start over
        if (prevCycles.length < 1)
            throw new Error('Got empty previous cycles');
        prevCycles.sort((a, b) => (a.counter > b.counter ? -1 : 1));
        // Add prevCycles to our cycle chain
        // commented out because not currently used
        // const combineCycles = []
        for (const prevCycle of prevCycles) {
            // Stop saving prevCycles if one of them is invalid
            if ((0, Cycles_1.validateCycle)(prevCycle, savedCycleRecord) === false) {
                Logger.mainLogger.error(`Record ${prevCycle.counter} failed validation`);
                Logger.mainLogger.debug('fail', prevCycle, savedCycleRecord);
                break;
            }
            savedCycleRecord = prevCycle;
            Logger.mainLogger.debug('Inserting archived cycle for counter', prevCycle.counter);
            const archivedCycle = createArchivedCycle(prevCycle);
            await Storage.insertArchivedCycle(archivedCycle);
        }
        endCycle = nextEnd - 1;
    }
}
exports.syncCyclesAndNodeList = syncCyclesAndNodeList;
function createArchivedCycle(cycleRecord) {
    const archivedCycle = {
        _id: '',
        cycleRecord: cycleRecord,
        cycleMarker: cycleRecord.marker,
        data: {},
        receipt: {},
        summary: {},
    };
    return archivedCycle;
}
async function downloadArchivedCycles(archiver, cycleToSyncTo, startCycle = 0) {
    let complete = false;
    let lastData = startCycle;
    let collector = [];
    let count = 0;
    const maxCount = Math.ceil((cycleToSyncTo - startCycle) / 5);
    while (!complete && count < maxCount) {
        Logger.mainLogger.debug(`Downloading archive from cycle ${lastData} to cycle ${lastData + 5}`);
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        const response = await P2P.getJson(`http://${archiver.ip}:${archiver.port}/full-archive?start=${lastData}&end=${lastData + 5}`);
        if (response && response.archivedCycles) {
            collector = collector.concat(response.archivedCycles);
            if (response.archivedCycles.length < 5) {
                complete = true;
                Logger.mainLogger.debug('Download completed');
            }
        }
        else {
            Logger.mainLogger.debug('Invalid download response');
        }
        count += 1;
        lastData += 5;
    }
    Logger.mainLogger.debug(`Downloaded archived cycles`, collector.length);
    return collector;
}
async function compareWithOldCyclesData(archiver, lastCycleCounter = 0) {
    let downloadedCycles;
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    const response = await P2P.getJson(`http://${archiver.ip}:${archiver.port}/cycleinfo?start=${lastCycleCounter - 10}&end=${lastCycleCounter - 1}`);
    if (response && response.cycleInfo) {
        downloadedCycles = response.cycleInfo;
    }
    else {
        throw Error(`Can't fetch data from cycle ${lastCycleCounter - 10} to cycle ${lastCycleCounter - 1}  from archiver ${archiver}`);
    }
    const oldCycles = await Storage.queryCycleRecordsBetween(lastCycleCounter - 10, lastCycleCounter + 1);
    downloadedCycles.sort((a, b) => (a.counter > b.counter ? 1 : -1));
    oldCycles.sort((a, b) => (a.counter > b.counter ? 1 : -1));
    let success = false;
    let cycle = 0;
    for (let i = 0; i < downloadedCycles.length; i++) {
        // eslint-disable-next-line security/detect-object-injection
        const downloadedCycle = downloadedCycles[i];
        // eslint-disable-next-line security/detect-object-injection
        const oldCycle = oldCycles[i];
        console.log(downloadedCycle, oldCycle);
        if (lib_types_2.Utils.safeStringify(downloadedCycle) !== lib_types_2.Utils.safeStringify(oldCycle)) {
            return {
                success,
                cycle,
            };
        }
        success = true;
        cycle = downloadedCycle.counter;
    }
    return { success, cycle };
}
exports.compareWithOldCyclesData = compareWithOldCyclesData;
async function syncStateMetaData(activeArchivers) {
    const randomArchiver = Utils.getRandomItemFromArr(activeArchivers)[0];
    const allCycleRecords = await Storage.queryAllCycleRecords();
    const lastCycleCounter = allCycleRecords[0].counter;
    const downloadedArchivedCycles = await downloadArchivedCycles(randomArchiver, lastCycleCounter);
    const networkReceiptHashesFromRecords = new Map();
    const networkDataHashesFromRecords = new Map();
    const networkSummaryHashesFromRecords = new Map();
    allCycleRecords.forEach((cycleRecord) => {
        if (cycleRecord.networkReceiptHash.length > 0) {
            cycleRecord.networkReceiptHash.forEach((hash) => {
                networkReceiptHashesFromRecords.set(hash.cycle, hash.hash);
            });
        }
        if (cycleRecord.networkDataHash.length > 0) {
            cycleRecord.networkDataHash.forEach((hash) => {
                networkDataHashesFromRecords.set(hash.cycle, hash.hash);
            });
        }
        if (cycleRecord.networkSummaryHash.length > 0) {
            cycleRecord.networkSummaryHash.forEach((hash) => {
                networkSummaryHashesFromRecords.set(hash.cycle, hash.hash);
            });
        }
    });
    for (let i = 0; i < downloadedArchivedCycles.length; i++) {
        // eslint-disable-next-line security/detect-object-injection
        const marker = downloadedArchivedCycles[i].cycleRecord.marker;
        // eslint-disable-next-line security/detect-object-injection
        const counter = downloadedArchivedCycles[i].cycleRecord.counter;
        // eslint-disable-next-line security/detect-object-injection
        const downloadedArchivedCycle = downloadedArchivedCycles[i];
        if (!downloadedArchivedCycle) {
            Logger.mainLogger.debug('Unable to download archivedCycle for counter', counter);
            continue;
        }
        let isDataSynced = false;
        let isReceiptSynced = false;
        let isSummarySynced = false;
        // Check and store data hashes
        if (downloadedArchivedCycle.data) {
            const downloadedNetworkDataHash = downloadedArchivedCycle.data.networkHash;
            const calculatedDataHash = (0, exports.calculateNetworkHash)(downloadedArchivedCycle.data.partitionHashes);
            if (downloadedNetworkDataHash === calculatedDataHash) {
                if (downloadedNetworkDataHash !== networkDataHashesFromRecords.get(counter)) {
                    Logger.mainLogger.debug('Different with hash from downloaded Cycle Records', 'state data', counter);
                }
                await Storage.updateArchivedCycle(marker, 'data', downloadedArchivedCycle.data);
                isDataSynced = true;
            }
            else {
                Logger.mainLogger.error('Different network data hash for cycle', counter);
            }
        }
        else {
            Logger.mainLogger.error(`ArchivedCycle ${downloadedArchivedCycle.cycleRecord.counter}, ${downloadedArchivedCycle.cycleMarker} does not have data field`);
        }
        // Check and store receipt hashes + receiptMap
        if (downloadedArchivedCycle.receipt) {
            // TODO: calcuate the network hash by hashing downloaded receipt Map instead of using downloadedNetworkReceiptHash
            const downloadedNetworkReceiptHash = downloadedArchivedCycle.receipt.networkHash;
            const calculatedReceiptHash = (0, exports.calculateNetworkHash)(downloadedArchivedCycle.receipt.partitionHashes);
            if (downloadedNetworkReceiptHash === networkReceiptHashesFromRecords.get(counter) ||
                downloadedNetworkReceiptHash === calculatedReceiptHash) {
                if (downloadedNetworkReceiptHash !== networkReceiptHashesFromRecords.get(counter)) {
                    Logger.mainLogger.debug('Different with hash from downloaded Cycle Records', 'receipt', counter);
                }
                await Storage.updateArchivedCycle(marker, 'receipt', downloadedArchivedCycle.receipt);
                isReceiptSynced = true;
            }
            else {
                Logger.mainLogger.error('Different network receipt hash for cycle', counter);
            }
        }
        else {
            Logger.mainLogger.error(`ArchivedCycle ${downloadedArchivedCycle.cycleRecord.counter}, ${downloadedArchivedCycle.cycleMarker} does not have receipt field`);
        }
        // Check and store summary hashes
        if (downloadedArchivedCycle.summary) {
            // TODO: calcuate the network hash by hashing downloaded summary Blobs instead of using downloadedNetworkSummaryHash
            const downloadedNetworkSummaryHash = downloadedArchivedCycle.summary.networkHash;
            const calculatedSummaryHash = (0, exports.calculateNetworkHash)(downloadedArchivedCycle.summary.partitionHashes);
            if (downloadedNetworkSummaryHash === calculatedSummaryHash) {
                if (downloadedNetworkSummaryHash !== networkSummaryHashesFromRecords.get(counter)) {
                    Logger.mainLogger.debug('Different with hash from downloaded Cycle Records', 'summary', counter);
                }
                await Storage.updateArchivedCycle(marker, 'summary', downloadedArchivedCycle.summary);
                isSummarySynced = true;
            }
            else {
                Logger.mainLogger.error('Different network summary hash for cycle', counter);
            }
        }
        else {
            Logger.mainLogger.error(`ArchivedCycle ${downloadedArchivedCycle.cycleRecord.counter}, ${downloadedArchivedCycle.cycleMarker} does not have summary field`);
        }
        if (isDataSynced && isReceiptSynced && isSummarySynced) {
            Logger.mainLogger.debug(`Successfully synced statemetadata for counter ${counter}`);
            if (counter > Cycles.lastProcessedMetaData) {
                Cycles.setLastProcessedMetaDataCounter(counter);
            }
        }
    }
    return false;
}
exports.syncStateMetaData = syncStateMetaData;
const calculateNetworkHash = (data) => {
    let hashArray = [];
    if (data) {
        for (const hash of Object.values(data)) {
            hashArray.push(hash);
        }
    }
    hashArray = hashArray.sort();
    const calculatedHash = Crypto.hashObj(hashArray);
    return calculatedHash;
};
exports.calculateNetworkHash = calculateNetworkHash;
async function queryDataFromNode(consensorNode, dataQuery, validateFn) {
    const request = {
        ...dataQuery,
        nodeInfo: State.getNodeInfo(),
    };
    let result = {
        success: false,
        completed: [],
        failed: [],
        covered: [],
        blobs: {},
    };
    try {
        const response = (await P2P.postJson(`http://${consensorNode.ip}:${consensorNode.port}/querydata`, request));
        if (response && request.type === 'RECEIPT_MAP') {
            const receiptMapData = response.data;
            // for (let counter in response.data) {
            result = await validateAndStoreReceiptMaps(receiptMapData, validateFn);
            // }
        }
        else if (response && request.type === 'SUMMARY_BLOB') {
            // for (let counter in response.data) {
            const summaryBlobData = response.data;
            result = await validateAndStoreSummaryBlobs(Object.values(summaryBlobData), validateFn);
            // }
        }
        return result;
    }
    catch (e) {
        Logger.mainLogger.error(e);
        Logger.mainLogger.error(`Unable to query complete querying ${request.type} from node`, consensorNode);
        return result;
    }
}
async function validateAndStoreReceiptMaps(receiptMapResultsForCycles, validateFn) {
    const completed = [];
    const failed = [];
    const coveredPartitions = [];
    const receiptMaps = {};
    for (const counter in receiptMapResultsForCycles) {
        const receiptMapResults = 
        // eslint-disable-next-line security/detect-object-injection
        receiptMapResultsForCycles[counter];
        for (const receiptMap of receiptMapResults) {
            const { cycle, partition } = receiptMap;
            if (validateFn) {
                const shouldProcess = validateFn(cycle, partition);
                if (!shouldProcess) {
                    continue;
                }
            }
            coveredPartitions.push(partition);
            const reciptMapHash = await Storage.queryReceiptMapHash(parseInt(counter), partition);
            if (!reciptMapHash) {
                Logger.mainLogger.error(`Unable to find receipt hash for counter ${counter}, partition ${partition}`);
                continue;
            }
            const calculatedReceiptMapHash = Crypto.hashObj(receiptMap);
            if (calculatedReceiptMapHash === reciptMapHash) {
                await Storage.updateReceiptMap(receiptMap);
                completed.push(partition);
                // eslint-disable-next-line security/detect-object-injection
                receiptMaps[partition] = receiptMap;
            }
            else {
                Logger.mainLogger.error(`Different hash while downloading receipt maps for counter ${counter}, partition ${partition}`);
                failed.push(partition);
            }
        }
    }
    return {
        success: true,
        completed,
        failed,
        covered: coveredPartitions,
        blobs: receiptMaps,
    };
}
async function validateAndStoreSummaryBlobs(statsClumpForCycles, validateFn) {
    const completed = [];
    const failed = [];
    const coveredPartitions = [];
    const blobs = {};
    for (const statsClump of statsClumpForCycles) {
        const { cycle, dataStats, txStats, covered } = statsClump;
        if (!covered)
            continue;
        for (const partition of covered) {
            if (validateFn) {
                const shouldProcess = validateFn(cycle, partition);
                if (!shouldProcess) {
                    continue;
                }
            }
            coveredPartitions.push(partition);
            let summaryBlob;
            const dataBlob = dataStats.find((d) => d.partition === partition);
            const txBlob = txStats.find((t) => t.partition === partition);
            const summaryHash = await Storage.querySummaryHash(cycle, partition);
            if (!summaryHash) {
                continue;
            }
            const summaryObj = {
                dataStats: dataBlob ? dataBlob.opaqueBlob : {},
                txStats: txBlob ? txBlob.opaqueBlob : {},
            };
            const calculatedSummaryHash = Crypto.hashObj(summaryObj);
            if (summaryHash !== calculatedSummaryHash) {
                failed.push(partition);
                continue;
            }
            if (dataBlob) {
                summaryBlob = {
                    ...dataBlob,
                };
            }
            if (txBlob) {
                if (!summaryBlob) {
                    summaryBlob = {
                        ...txBlob,
                    };
                }
                else if (summaryBlob) {
                    summaryBlob.latestCycle = txBlob.latestCycle;
                    summaryBlob.opaqueBlob = {
                        ...summaryBlob.opaqueBlob,
                        ...txBlob.opaqueBlob,
                    };
                }
            }
            if (summaryBlob) {
                try {
                    await Storage.updateSummaryBlob(summaryBlob, cycle);
                    completed.push(partition);
                    // eslint-disable-next-line security/detect-object-injection
                    blobs[partition] = summaryBlob;
                }
                catch (e) {
                    Logger.mainLogger.error('Unable to store summary blob', e);
                    throw new Error('Unable to store summary blob');
                }
            }
        }
    }
    return {
        success: true,
        completed,
        failed,
        covered: coveredPartitions,
        blobs,
    };
}
exports.emitter.on('selectNewDataSender', async (newSenderInfo, taggedDataRequest) => {
    //let request = {
    //...dataRequest,
    //nodeInfo: State.getNodeInfo()
    //}
    if (exports.socketClients.has(newSenderInfo.publicKey)) {
        const response = await P2P.postJson(`http://${newSenderInfo.ip}:${newSenderInfo.port}/requestdata`, taggedDataRequest);
        Logger.mainLogger.debug('/requestdata response', response);
    }
});
//# sourceMappingURL=StateMetaData.js.map