"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getLatestCycleRecordsFromCache = exports.addCyclesToCache = void 0;
const Config_1 = require("../Config");
const cycles_1 = require("../dbstore/cycles");
const Crypto = require("../Crypto");
let cachedCycleRecords = [];
const signedCacheCycleRecords = new Map();
let lastCacheUpdateFromDBRunning = false;
async function updateCacheFromDB() {
    if (lastCacheUpdateFromDBRunning) {
        return;
    }
    lastCacheUpdateFromDBRunning = true;
    try {
        cachedCycleRecords = await (0, cycles_1.queryLatestCycleRecords)(Config_1.config.REQUEST_LIMIT.MAX_CYCLES_PER_REQUEST);
    }
    catch (error) {
        console.log('Error updating latest cache: ', error);
    }
    finally {
        lastCacheUpdateFromDBRunning = false;
    }
}
async function addCyclesToCache(cycles) {
    if (cachedCycleRecords.length === 0) {
        await updateCacheFromDB();
    }
    for (const cycle of cycles) {
        cachedCycleRecords.unshift(cycle);
    }
    cycles.sort((a, b) => a.counter - b.counter);
    if (cachedCycleRecords.length > Config_1.config.REQUEST_LIMIT.MAX_CYCLES_PER_REQUEST) {
        cachedCycleRecords.splice(Config_1.config.REQUEST_LIMIT.MAX_CYCLES_PER_REQUEST);
    }
    signedCacheCycleRecords.clear();
}
exports.addCyclesToCache = addCyclesToCache;
async function getLatestCycleRecordsFromCache(count) {
    if (cachedCycleRecords.length === 0) {
        await updateCacheFromDB();
    }
    if (signedCacheCycleRecords.has(count))
        return signedCacheCycleRecords.get(count);
    const cycleInfo = cachedCycleRecords.slice(0, count);
    const signedCycleRecords = Crypto.sign({ cycleInfo });
    signedCacheCycleRecords.set(count, signedCycleRecords);
    return signedCycleRecords;
}
exports.getLatestCycleRecordsFromCache = getLatestCycleRecordsFromCache;
//# sourceMappingURL=cycleRecordsCache.js.map