"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.initAccounts = void 0;
const SchemaHelpers_1 = require("../../utils/serialization/SchemaHelpers");
const AJVSchemaEnum_1 = require("../enum/AJVSchemaEnum");
// Define the schema for AccountsCopy
const schemaAccountsCopy = {
    type: 'object',
    properties: {
        accountId: { type: 'string' },
        data: { type: 'object', additionalProperties: true },
        timestamp: { type: 'integer', minimum: 0 },
        hash: { type: 'string' },
        cycleNumber: { type: 'integer', nullable: true },
        isGlobal: { type: 'boolean' }
    },
    required: ['accountId', 'data', 'timestamp', 'hash', 'isGlobal'] // cycleNumber is optional
};
// Define the schema for DbAccountCopy
const schemaDbAccountCopy = {
    type: 'object',
    properties: {
        ...schemaAccountsCopy.properties,
        data: { type: 'string' } // Overriding the `data` field to be a string in DbAccountCopy
    },
    required: ['accountId', 'data', 'timestamp', 'hash', 'isGlobal'] // Required fields remain the same
};
// Function to initialize schemas
function initAccounts() {
    addSchemaDependencies();
    addSchemas();
}
exports.initAccounts = initAccounts;
// Function to add schema dependencies
function addSchemaDependencies() {
    // No external dependencies
}
// Function to register schemas
function addSchemas() {
    (0, SchemaHelpers_1.addSchema)(AJVSchemaEnum_1.AJVSchemaEnum.AccountsCopy, schemaAccountsCopy);
}
//# sourceMappingURL=Accounts.js.map